#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Generate spherical radial distribution function from a PDB

Created on Wed Dec  9 11:11:32 2020

@author: Natalie Krzyzanowski
Translated from Mathematica script written by Dr. Ursula Perez-Salas
Lines 57 - 77 are taken from the program written by:
Copyright (C) 2011 Stas Bevc, stas.bevc@cmm.ki.si, www.sicmm.org
"""

'''CHANGE PARAMETERS AT THE BOTTOM OF THE SCRIPT'''

'''
Import all necessary packages
'''
from os import listdir, path
from time import time

from math import sqrt, pi

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


'''
Here we define the g(r) function, which takes as input:
    pdbsdir: string, the directory containing the desired pdbs
    maxbin: int, the maximum number of bins in the histogram which defines the arclength r*theta
    ref: int, the number of reference atoms to use, between 1 and max (usually 1260)
    choose: int, 1 to use all atoms as references, 0 for the # chosen in ref above 
    
    The function returns no outputs, produces a plot of g(r) vs r*theta,
    and exports r*theta, g(r) columns into an excel file in the pdbsdir directory. 
'''

def gofr(pdbsdir, ref,choose,step):
    pdbs = listdir(pdbsdir) #load each pdb
    count = 0               #number of pdbs
    nstep = len(pdbs)
    start_time = time()
    '''
    Pandas DataFrame holds the following variables:
        i is the bin length and is valued from 1 to 1058
        sum_binCount is the bin count sum from 1 to i 
        binCountTransform is sum_binCount normalized using:
            binCountTransform = sum_binCount / expr
            where expr = (2*pi*radii[n]^2 * (1-Cos(1/radii[n])) * N/(4*pi*radii[n]^2) * (N-1)/N))
            reference: "Introduction to Modern Statistical Mechanics," David Chandler, pg. 222
    '''

    gofrs = pd.DataFrame()
    
    #loop over pdb files
    for pdb in pdbs:
        if not pdb.endswith(".pdb"):
            nstep -= 1
            continue # skip other files
        count += 1

        '''
        Import the PDB coordinates, which could be 1 phase
        or the whole vesicle.
        '''
        atoms = []
        lines = open(pdbsdir+"/"+pdb)
        for line in lines:
            if line[0:4] != "ATOM":
                continue # skip other lines
            coords = list(map(float, [line[30:38],line[38:46],line[46:54]]))

            atoms.append((coords[0], coords[1], coords[2]))

        #all atoms in pdb
        npart = len(atoms)

        '''
        1) Fix an atom in the pdb
        2) Calculate g(r) of all remaining atoms relative to it
        3) Step foward in a loop for the next atom
        4) Repeat 1-3 until all desired (ref or max) atoms have been fixed.
        '''
            
        radiiFixed = []  #fixed atom radii
        radiiFree = []   #free atom radii

        '''
        Sometimes we want to pick just one or a few atoms to fix 
        (fixed atoms are also called reference atoms here). 
            ref denotes the number of desired reference atoms
            npart denotes all atoms in pdb 
        '''
        if choose == 1:         #for a [ref] number of fixed atoms
            ref = ref
        if choose == 0:         #for all atoms in the pdb
            ref = npart
            
        for i in range(0,ref):     
            ax = float(atoms[i][0])
            ay = float(atoms[i][1])
            az = float(atoms[i][2])
            
            if i == ref/4:
                print("A quarter of the way through the fixed atom list...")
            if i == ref/2:
                print("Halfway through the fixed atom list...")
            if i == ref/4 * 3:
                print("3/4 of the way done, almost there...")
            
            '''
            a is the equivalent of radii[n] 
            '''
            
            a = (sqrt(ax**2+ay**2+az**2))
            radiiFixed.append(a)
            
            '''
            Find coordinate of n+1 atom 
            and calculate arc length between n, fixed atom
            and all the remaining atoms
            '''
            
            arcLength = []
            
          # for j in range(i+1,npart): #don't repeat previous pairs
            for j in range(0,npart): #repeat previous pairs 
                bx = float(atoms[j][0])
                by = float(atoms[j][1])
                bz = float(atoms[j][2])
                
                b = (sqrt(bx**2+by**2+bz**2))
                radiiFree.append(b)
                
                #dot product 
                dot = ((ax*bx+ay*by+az*bz))
                
                '''
                Calculate s, the arc length, and store it in a list.
                S is calculated between each of the free atoms relative to
                the current fixed atom
                '''
                
                s = a * np.arccos(dot/(a*b))
                arcLength.append(s)

                        
            maxbin = int(round(max(radiiFree) * pi))  #maximum bin is the max arclength

            '''
            Now bin all arclengths between 1 and maxBin (1058)
            
            elements are counted in bin i if b_i <= x < b_(i+1)
            
            expected output: s_bin < 10 for any bin
            
            returns binCount and bin_edges
            binCount a numpy ndarray holding int counts
            bin_edges is a numpy hdarray holding bin edges
            '''
            
            binCount, bin_edges  = np.histogram(arcLength,bins=maxbin,range=(0,maxbin))
            kList = []                   #this will become the r's column in the file
            binCountTransformList = []   #create a list of length maxbin, for each fixed atom
            
            '''
            Below we calculate g(r) vs r*theta
            
            Transform binCount into binCountTransform and 
            put into DataFrame as a new Series
            '''

            for k in range(1,maxbin+1-step):
                kList.append(k)
                sumBinCount = 0    
                
                #sum counts from bin i to bin i + step 
                for z in range(k,k+step): #Sum[bins3[[j]], {j, i, i+step}])
                    sumBinCount += binCount[z]    
                
                binCountTransform = sumBinCount / (2. * pi * radiiFixed[i]**2)
                binCountTransform /= (np.cos(k/radiiFixed[i])-np.cos((k+step)/radiiFixed[i]))
                binCountTransform /= (float(npart)/(4. * pi * radiiFixed[i]**2))
                binCountTransform /= (float(npart-1)/float(npart))
                
                binCountTransformList.append(binCountTransform)
                
            if i == 0:
                gofrs['i'] = kList
        
            '''
            Keep track of every transformed sum for all fixed atoms
            '''
            name = "binCountTransform" + str(i)
            gofrs[name] = binCountTransformList
            
            
        filename = path.basename(pdbsdir+"/"+pdb)
        filename = filename.rstrip('.pdb')
        
    '''
    After looping through all atom pairs, we can now average
    the binCountTransform to get g(r) 

    First, drop the i column so it doesn't get included in the averaging
    '''
    
    gofrs2 = gofrs.drop(['i'],axis=1)

    #Create a new column with the mean, which is the g(r)
    gofrs['mean'] = gofrs2.mean(axis=1)
    
    #Grab only the r and g(r)
    gofrs = gofrs.loc[:,['i','mean']]
    
    
    '''
    Plot
    '''
    
    gofrs.plot(kind='scatter',
               x='i',
               y='mean',
               color='red',
               grid=True,
               ylim=(0,2.0))
    

    plt.savefig(pdbsdir+'/gofrs_' + filename + '.png')
    #plt.show()
     
    
    '''
    Export into excel file containing the r's and the g(r)'s
    '''

    gofrs.to_csv(pdbsdir + '/gofrs_' + filename + '.csv',
                 index = False, header=True)

    end_time = time()
    total_time = end_time - start_time
    
    print("This program took %0.2f seconds to calculate %d PDBs"
          %(total_time,count))

    lines.close()

'''
Initialize the directory, maxbin, ref, and choose
Then call the function
'''

'''CHANGE PARAMETERS BELOW:
    pdbsdir is the file directory containing the PDB with the phase of interest (liquidus or solidus). If more than one PDB this script will average all into one outputfile.
    maxbin is the max arc length
    ref is number of reference atoms to calculate g(r) relative to
    choose is 1 or 0, where 1 means you use the number of reference
        atoms given above, and 0 means you want to use the entire PDB length
'''

#pdbsdir="."        #current directory the script is in
pdbsdir = "/Users/Natalie/Thesis/gofr_script/other/test"  #Macbook
ref = 1
step = 50
choose = 0  # 1 for ref atom, 0 for all atoms 

#Call the g(r) function
gofr(pdbsdir,ref,choose,step)
